"""Interface to the catalog of SXS data

"""

from .catalog import Catalog
from .create import _create
from .description import catalog_file_description


formats = {
    None: Catalog,
    "": Catalog,
}



"""Functions to compare SXS and LVC-NR formatted data sets"""


def compare_attributes(lvc, sxs_metadata, count_errors):
    """Compares LVC attributes to SXS metadata"""
    import numpy as np
    from .metadata import sxs_id_from_alt_names, msun_seconds

    def compare_attribute(lvc_key, sxs_value):
        lvc_value = lvc.attrs[lvc_key]
        if lvc_value==sxs_value or np.allclose(lvc_value, sxs_value, atol=1e-15, rtol=1e-15, equal_nan=True):
            count_errors("[=] {0} ({1})".format(lvc_key, lvc_value))
        else:
            count_errors("[x] {0} (lvc: {1}, sxs: {2})".format(lvc_key, lvc_value, sxs_value))

    count_errors("# Comparing metadata")
    sxs_id = sxs_id_from_alt_names(sxs_metadata["alternative_names"])
    compare_attribute("name", sxs_id)

    mass1 =  sxs_metadata["reference_mass1"]
    mass2 =  sxs_metadata["reference_mass2"]
    compare_attribute("mass1", mass1)
    compare_attribute("mass2", mass2)

    eta = (mass1 * mass2) / (mass1 + mass2)**2
    compare_attribute("eta", eta)

    compare_attribute("spin1x", sxs_metadata["reference_dimensionless_spin1"][0])
    compare_attribute("spin1y", sxs_metadata["reference_dimensionless_spin1"][1])
    compare_attribute("spin1z", sxs_metadata["reference_dimensionless_spin1"][2])
    compare_attribute("spin2x", sxs_metadata["reference_dimensionless_spin2"][0])
    compare_attribute("spin2y", sxs_metadata["reference_dimensionless_spin2"][1])
    compare_attribute("spin2z", sxs_metadata["reference_dimensionless_spin2"][2])

    omega_orbit_vec = sxs_metadata["reference_orbital_frequency"]
    omega_orbit = np.linalg.norm(omega_orbit_vec)
    compare_attribute("Omega", omega_orbit)

    f_lower_at_1_msun = 2.0 * omega_orbit / (2.0 * np.pi * msun_seconds)
    compare_attribute("f_lower_at_1MSUN", f_lower_at_1_msun)

    eccentricity = str(sxs_metadata['reference_eccentricity'])
    if eccentricity == '[simulation too short]' or eccentricity == '[unknown]':
        eccentricity = -1.0
    elif eccentricity[0] == "<" or eccentricity[0] == ">":
        eccentricity = float(eccentricity[1:])
    else:
        eccentricity = float(eccentricity)
    compare_attribute("eccentricity", eccentricity)

    mean_anomaly = sxs_metadata['reference_mean_anomaly']
    if mean_anomaly == '[unknown]':
        mean_anomaly = -1.0
    else:
        mean_anomaly = float(mean_anomaly)
    compare_attribute("mean_anomaly", mean_anomaly)

    compare_attribute("Omega", omega_orbit)


def compare_waveform_splines(lvc, sxs_waveform, mode_string, extrap_order="Extrapolated_N2"):
    import numpy as np
    from . import Dataset

    start_time = lvc.attrs["NR_start_time"]
    peak_time = lvc.attrs["NR_peak_time"]

    # Get SXS values
    sxs_key = "Y" + mode_string + ".dat"
    hlm = sxs_waveform[extrap_order + ".dir"][sxs_key]
    times_raw = hlm[:, 0]
    start_h = np.abs(times_raw - start_time).argmin() + 1
    sxs_values = hlm[start_h:, 1] + 1j * hlm[start_h:, 2]
    sxs_times = hlm[start_h:, 0] - peak_time

    # Get LVC values
    lvc_amp = Dataset.read(lvc["amp"+mode_string]).spline(sxs_times)
    lvc_phase = Dataset.read(lvc["phase"+mode_string]).spline(sxs_times)
    lvc_values = lvc_amp * np.exp(1j * lvc_phase)

    # Linf norm of difference
    diff = np.max(np.abs(lvc_values - sxs_values))
    return diff


def compare_horizon_splines(lvc, sxs_horizons, key, count_errors):
    import numpy as np
    from . import Dataset

    start_time = lvc.attrs["NR_start_time"]
    peak_time = lvc.attrs["NR_peak_time"]

    if "remnant" in key:
        ah_key = "AhC.dir"
    elif "2" in key:
        ah_key = "AhB.dir"
    elif "1" in key:
        ah_key = "AhA.dir"
    else:
        count_errors('# Unknown key: {0}'.format(key))
        return np.inf

    if "mass" in key:
        quantity_key = "ChristodoulouMass"
        component_index = 1
    elif "spin" in key:
        quantity_key = "chiInertial"
        if 'x' in key:
            component_index = 1
        elif 'y' in key:
            component_index = 2
        elif 'z' in key:
            component_index = 3
    elif "position" in key:
        quantity_key = "CoordCenterInertial"
        if 'x' in key:
            component_index = 1
        elif 'y' in key:
            component_index = 2
        elif 'z' in key:
            component_index = 3
    quantity = sxs_horizons[ah_key][quantity_key + ".dat"]
    times_raw_ah = quantity[:, 0]
    start_ah = np.argmin(np.abs(times_raw_ah - start_time))
    sxs_values = quantity[start_ah:, component_index]
    sxs_times = quantity[start_ah:, 0] - peak_time

    # Linf norm of difference
    lvc_values = Dataset.read(lvc[key]).spline(sxs_times)
    diff = np.max(np.abs(lvc_values - sxs_values))
    return diff


def compare_time_series(lvc, sxs_horizons, key, extrap="Extrapolated_N2"):
    import numpy as np

    lvc_times = np.array(lvc[key])
    start_time = lvc.attrs["NR_start_time"]
    peak_time = lvc.attrs["NR_peak_time"]
    if "HorizonA" in key:
        times_raw_ah = sxs_horizons["AhA.dir"]["ArealMass.dat"][:, 0]
        start_ah = np.argmin(np.abs(times_raw_ah - start_time))
        sxs_times = times_raw_ah[start_ah:] - peak_time
    elif "HorizonB" in key:
        times_raw_ah = sxs_horizons["AhB.dir"]["ArealMass.dat"][:, 0]
        start_ah = np.argmin(np.abs(times_raw_ah - start_time))
        sxs_times = times_raw_ah[start_ah:] - peak_time
    elif "CommonHorizon" in key:
        times_raw_ah = sxs_horizons["AhC.dir"]["ArealMass.dat"][:, 0]
        start_ah = np.argmin(np.abs(times_raw_ah - start_time))
        sxs_times = times_raw_ah[start_ah:] - peak_time

    # Linf norm of difference
    diff = np.max(np.abs(lvc_times - sxs_times))
    return diff


def compare_wave_time_series(lvc, sxs_waveform, count_errors, extrap="Extrapolated_N2"):
    import numpy as np

    lvc_wave_times = np.array(lvc["NRtimes"])
    start_time = lvc.attrs["NR_start_time"]
    peak_time = lvc.attrs["NR_peak_time"]

    # Check that the time series is as expected
    sxs_key = "Y_l2_m2.dat"
    hlm = sxs_waveform[extrap + ".dir"][sxs_key]
    times_raw = hlm[:, 0]
    start_h = np.abs(times_raw - start_time).argmin()
    sxs_times = hlm[start_h:, 0] - peak_time

    diff = np.max(np.abs(sxs_times - lvc_wave_times))
    if diff != 0.0:
        count_errors("[x] NRtimes differs from SXS l=m=2 times (diff = {0})".format(diff))
    else:
        count_errors("[=] NRtimes agrees with SXS l=m=2 times (diff = {0})".format(diff))


def compare_peaks(lvc, sxs_waveform, lvc_amp_keys, count_errors, extrap_order="Extrapolated_N2"):
    import numpy as np
    from . import Dataset

    start_time = lvc.attrs["NR_start_time"]
    peak_time = lvc.attrs["NR_peak_time"]

    # Get SXS time array
    sxs_key = "Y_l2_m2.dat"
    hlm = sxs_waveform[extrap_order + ".dir"][sxs_key]
    times_raw = hlm[:, 0]
    start_h = np.abs(times_raw - start_time).argmin() + 1
    sxs_times = hlm[start_h:, 0] - peak_time

    # Get LVC values
    lvc_l2norm = np.zeros_like(sxs_times)
    for lvc_amp_key in lvc_amp_keys:
        lvc_l2norm += Dataset.read(lvc[lvc_amp_key]).spline(sxs_times)**2

    # Check that the peak LVC amplitude is near t=0
    lvc_t_peak = sxs_times[np.argmax(lvc_l2norm)]
    if not np.allclose(lvc_t_peak, 0.0, atol=1e-12, rtol=0.0):
        count_errors("[x] LVC peak amplitude does not occur near t=0 (occurs at {0})".format(lvc_t_peak))
    else:
        count_errors("[=] LVC peak amplitude occurs near t=0 (occurs at {0})".format(lvc_t_peak))


def compare(lvc_file, sxs_data_path, verbosity=1):
    """Compare an LVC-format data file to SXS-format files

    Parameters
    ----------
    lvc_file : str
        Path to SXS_BBH_####_Res#.h5 file
    sxs_data_path : str
        Path to directory containing rhOverM_Asymptotic_GeometricUnits_CoM.h5,
        Horizons.h5, and metadata.json files.
    verbosity : int, optional
        If 0, don't print anything ever; if 1 (the default), print only if errors
        were found; if greater than 1, print full results.

    """
    import os
    import json
    import numpy as np
    import h5py

    class ErrorCounter(object):
        def __init__(self, verbosity):
            self.verbosity = verbosity
            self.errors = 0
            self.log = ''

        def __call__(self, string):
            if string.startswith('[x]'):
                self.errors += 1
            if self.verbosity > 1:
                print(string)
            else:
                self.log += string + '\n'

        def __del__(self):
            if self.errors and self.verbosity==1:
                print(self.log)

    count_errors = ErrorCounter(verbosity)

    command = "sxs.format.lvc.compare.to_sxs({lvc_file!r}, {sxs_data_path!r}, verbosity={verbosity!r})""".format(
        lvc_file=lvc_file,
        sxs_data_path=sxs_data_path,
        verbosity=verbosity
    )
    count_errors("# " + command)

    sxs_waveform_file = os.path.join(sxs_data_path, 'rhOverM_Asymptotic_GeometricUnits_CoM.h5')
    sxs_horizons_file = os.path.join(sxs_data_path, 'Horizons.h5')
    sxs_metadata_file = os.path.join(sxs_data_path, 'metadata.json')

    with open(sxs_metadata_file, 'r') as m:
        sxs_metadata = json.load(m)

    with h5py.File(lvc_file, 'r') as lvc:

        compare_attributes(lvc, sxs_metadata, count_errors)

        # Divide lvc keys into groups
        lvc_keys = list(lvc)
        lvc_amp_keys = [key for key in lvc_keys if key.startswith('amp_l')]
        lvc_phase_keys = [key for key in lvc_keys if key.startswith('phase_l')]
        lvc_vs_time_keys = [key for key in lvc_keys if key.endswith('-vs-time')]
        lvc_horizon_time_keys = [key for key in lvc_keys if key in ['CommonHorizonTimes', 'HorizonATimes', 'HorizonBTimes']]

        # Check to make sure we understand exactly what keys are here
        lvc_ignored_keys = ['auxiliary-info']
        lvc_known_keys = set(lvc_amp_keys + lvc_phase_keys + ['NRtimes']
                             + lvc_vs_time_keys + lvc_horizon_time_keys + lvc_ignored_keys)
        if set(lvc_keys).symmetric_difference(lvc_known_keys):
            count_errors('[x] Unclassified keys:', set(lvc_keys).symmetric_difference(lvc_known_keys))

        # Ensure that waveform keys are properly matched
        amp_modes = [k[3:] for k in lvc_amp_keys]
        phase_modes = [k[5:] for k in lvc_phase_keys]
        if set(amp_modes).symmetric_difference(phase_modes):
            count_errors('[x] Mismatched amp and phase modes:', set(amp_modes).symmetric_difference(phase_modes))

        # Compare waveforms
        with h5py.File(sxs_waveform_file, 'r') as sxs_waveform:
            count_errors("# Comparing waveforms")
            compare_wave_time_series(lvc, sxs_waveform, count_errors)
            # compare_peaks(lvc, sxs_waveform, lvc_amp_keys, count_errors)
            for mode_string in amp_modes:
                diff = compare_waveform_splines(lvc, sxs_waveform, mode_string)
                tol = lvc['phase'+mode_string]['tol'][()] + lvc['amp'+mode_string]['tol'][()]
                result = '=' if np.abs(diff) < tol else 'x'
                count_errors("[{0}] Y{1} (diff = {2}, tol = {3})".format(result, mode_string, diff, tol))

        # Compare horizon quantities
        with h5py.File(sxs_horizons_file, 'r') as sxs_horizons:
            count_errors("# Comparing horizon splines")
            for key in lvc_vs_time_keys:
                if key.startswith('LNhat') or key.startswith('nhat') or key.startswith('Omega'):
                    count_errors("[?] {0}".format(key))
                else:
                    diff = compare_horizon_splines(lvc, sxs_horizons, key, count_errors)
                    tol = lvc[key]['tol'][()]
                    result = '=' if np.abs(diff) < tol else 'x'
                    count_errors("[{0}] {1} (diff = {2}, tol = {3})".format(result, key, diff, tol))

            count_errors("# Comparing time series")
            eps = 1.e-15
            for key in lvc_horizon_time_keys:
                try:
                    diff = compare_time_series(lvc, sxs_horizons, key)
                    result = '=' if diff < eps else 'x'
                    count_errors("[{0}] {1} (diff = {2})".format(result, key, diff))
                except:
                    count_errors("[x] Cannot diff time series {0}".format(key))

    return count_errors.errors



def suppressor(data, tolerance, pad=0):
    """Identify elements of data close to 0

    Parameters
    ----------
    data : array
        This can be either real or complex data.
    tolerance : float
        Data of absolute value below this tolerance will be set to zero.
    pad : int, optional
        The function will only suppress data that has at least `pad` points between
        it and the nearest value that is above the tolerance.  Defaults to 0.

    """
    import numpy as np
    # Start off suppressing everything below tolerance
    suppressed = np.abs(data) < tolerance
    if pad > 0:
        # Count how many neighbors (within `pad` points to either side) are also suppressed
        counter = np.ones(2*pad+1, dtype=int)
        count_neighboring_suppressions = np.convolve(suppressed, counter, mode='same')
        # Only suppress those whose neighbors are all suppressed
        suppressed = count_neighboring_suppressions == np.sum(counter)
    return suppressed


def suppress(data, tolerance, pad=0, inplace=True):
    """Set data close to 0 to exactly 0

    Parameters
    ----------
    data : array
        This can be either real or complex data.
    tolerance : float
        Data of absolute value below this tolerance will be set to zero.
    pad : int, optional
        The function will only suppress data that has at least `pad` points between
        it and the nearest value that is above the tolerance.  Defaults to 0.
    inplace : bool, optional
        If True (the default), overwrite the `data` array and return it; if False,
        copy it, suppress small numbers, and return the copy.

    """
    import numpy as np
    if inplace:
        data[suppressor(data, tolerance, pad=pad)] = 0.0
        output = data
    else:
        output = np.copy(data)
        output[suppressor(data, tolerance, pad=pad)] = 0.0
    return output



"""File I/O for SpEC-format Horizons.h5 files"""


def save(horizons, file):
    """Save `Horizons` object as SpEC Horizons.h5 file

    Parameters
    ----------
    horizons : sxs.Horizons
        Horizons object to be written to file.
    file : file-like object, string, or pathlib.Path
        Path to the file on disk or a file-like object (such as an open file
        handle) to be written by h5py.File.

    See Also
    --------
    sxs.horizons.spec_horizons_h5.load : load the output file format

    """
    import h5py

    with h5py.File(file, "w") as f:
        f.attrs["sxs_format"] = "horizons.spec_horizons_h5"
        for horizon_name in "ABC":
            horizon = horizons[horizon_name]
            if horizon is not None:
                g = f.create_group(f"Ah{horizon_name}.dir")
                for dat in [
                    "ArealMass.dat", "ChristodoulouMass.dat", "DimensionfulInertialSpinMag.dat", "chiMagInertial.dat",
                    "CoordCenterInertial.dat", "DimensionfulInertialSpin.dat", "chiInertial.dat"
                ]:
                    data = horizon[dat]
                    chunks = (data.shape[0], 1)
                    g.create_dataset(dat, data=data, shuffle=True, compression="gzip", chunks=chunks)


def load(file):
    """Load SpEC-format Horizons.h5 file as `Horizons` object

    Parameters
    ----------
    file : file-like object, string, or pathlib.Path
        Path to the file on disk or a file-like object (such as an open file
        handle) to be opened by h5py.File.

    Returns
    -------
    horizons : sxs.Horizons
        This is a container for the horizon objects.  See Notes below.

    See also
    --------
    sxs.Horizons : Container object for all of the horizons
    sxs.HorizonQuantities : Container objects for each of the horizons
    sxs.horizons.spec_horizons_h5.save : save to this file format

    Notes
    -----
    The returned object can be indexed just like the original SpEC-format
    Horizons.h5 file:

        horizons["AhA.dir/CoordCenterInertial.dat"]

    However, the `horizons` object also has a more natural and general interface
    that should be preferred for compatibility with other formats, in which the
    same vector-valued function of time can be accessed as

        horizons.A.coord_center_inertial

    See the documentation of `Horizons` and `HorizonQuantities` for more details.

    """
    import h5py
    from .. import Horizons, HorizonQuantities

    hqs = {}
    with h5py.File(file, "r") as f:
        sxs_format = f.attrs.get("sxs_format", "horizons.spec_horizons_h5")
        if sxs_format != "horizons.spec_horizons_h5":
            raise ValueError(
                f"\nAttribute 'sxs_format' found in '{file}' is '{sxs_format}'.\n"
                f"This function only accepts 'horizons.spec_horizons_h5' formats.\n"
                f"Use a higher-level `load` function to auto-detect the format."
            )
        for horizon_name in "ABC":
            dir_name = f"Ah{horizon_name}.dir"
            if dir_name in f:
                g = f[dir_name]
                horizon = HorizonQuantities(
                    time=g["ArealMass.dat"][:, 0],
                    areal_mass=g["ArealMass.dat"][:, 1],
                    christodoulou_mass=g["ChristodoulouMass.dat"][:, 1],
                    coord_center_inertial=g["CoordCenterInertial.dat"][:, 1:],
                    dimensionful_inertial_spin=g["DimensionfulInertialSpin.dat"][:, 1:],
                    chi_inertial=g["chiInertial.dat"][:, 1:],
                )
                hqs[horizon_name] = horizon
    return Horizons(**hqs)



"""A very simple function for fitting output to a console"""

def fit_to_console(obj, initial_indent='', subsequent_indent='', width=None,):
    """Return a string formatted to fit nicely in the console

    This is essentially a wrapper around textwrap and pprint; it pretty-prints the input object,
    then wraps it to the width of the output console.  The returned object is a string that can be
    printed without further adjustment.

    Parameters
    ----------
    obj: object
        Python object to be printed
    initial_indent: str [defaults to '']
        String that will be prepended to the first line of wrapped output.  Counts towards the
        line's width.  Note that this need not be whitespace; it could be a label, for example, to
        explain what `obj` is.
    subsequent_indent: str [defaults to '']
        String that will be prepended to all lines save the first of wrapped output; also counts
        towards each line's width.
    width: int or None [defaults to None]
        Full width of text to be output.  If None, the terminal size is detected, and the number of
        columns (minus 1) is used.

    """
    import shutil
    import textwrap
    import pprint
    columns, lines = shutil.get_terminal_size()
    full_width = width or columns-1
    wrapper = textwrap.TextWrapper(width=full_width, initial_indent=initial_indent, subsequent_indent=subsequent_indent)
    text = pprint.pformat(obj)
    return wrapper.fill(text)



"""Utility classes for lvcnr submodule"""

class Waveform(object):
    """Contains a waveform"""
    def __init__(self, *args, **kwargs):
        if args or kwargs:
            raise ValueError("This is an empty constructor; use `from_data` or `read` to add data.")

    @classmethod
    def read(cls, file_path, group_path='/', modes=None, ell_min=2, ell_max=8):
        import numpy as np
        import h5py
        w = cls()
        modes_string = "Y_l{0[0]}_m{0[1]}.dat"
        modes = modes or [[ell, m] for ell in range(ell_min, ell_max+1) for m in range(-ell, ell+1)]
        with h5py.File(file_path, 'r') as f:
            h = f[group_path]
            w.t = h[modes_string.format(modes[0])][:, 0]
            w.t.setflags(write=False)
            w.data = np.empty((w.t.size, len(modes)*2))
            for i, mode in enumerate(modes):
                w.data[:, 2*i:2*i+2] = h[modes_string.format(mode)][:, 1:3]
            w.data = w.data.view(complex)
            w.data.setflags(write=False)

            w.version_hist = f.get('VersionHist.ver', None)
            if w.version_hist is not None:
                w.version_hist = w.version_hist[:]
        return w

    def peak_index(self, start_index=0):
        import numpy as np
        return start_index + np.argmax(np.linalg.norm(self.data[start_index:], axis=1))


class WaveformAmpPhase(Waveform):
    @classmethod
    def read(cls, file_path, group_path='/', mode_strings=None, ell_min=2, ell_max=8, start_time=None):
        import numpy as np
        import h5py
        w = cls()
        mode_strings = mode_strings or ["Y_l{0}_m{1}.dat".format(ell, m) for ell in range(ell_min, ell_max+1) for m in range(-ell, ell+1)]
        with h5py.File(file_path, 'r') as f:
            h = f[group_path]
            t = h[mode_strings[0]][:, 0]
            if start_time is None:
                start_index = 0
            else:
                start_index = np.argmin(np.abs(t - start_time)) + 1
            w.t = t[start_index:].copy()
            w.t.setflags(write=False)
            w.data = np.empty((w.t.size, len(mode_strings)*2))
            d = np.empty((w.t.size, 2), dtype=float)
            c = d.view(complex).reshape(-1)
            for i, mode_string in enumerate(mode_strings):
                d[:, :] = h[mode_string][start_index:, 1:3]
                w.data[:, 2*i] = np.abs(c)
                w.data[:, 2*i+1] = np.unwrap(np.angle(c))
            w.data.setflags(write=False)
            w.version_hist = f.get('VersionHist.ver', None)
            if w.version_hist is not None:
                w.version_hist = w.version_hist[:]
        return w

    def peak_index(self, start_index=0):
        import numpy as np
        return start_index + np.argmax(np.linalg.norm(self.data[start_index:, ::2], axis=1))



"""The main container for waveform objects with mode weights"""

import re
import numpy as np
import quaternionic
import spherical
from .. import TimeSeries, jit
from . import WaveformMixin

NRAR_mode_regex = re.compile(r"""Y_l(?P<L>[0-9]+)_m(?P<M>[-+0-9]+)\.dat""")


class WaveformModes(WaveformMixin, TimeSeries):
    """Array-like object representing time-series data of SWSH modes

    We generally assume that the data represent mode weights in expansions of
    functions in terms of spin-weighted spherical harmonics (where the standard
    spherical harmonics just happen to have spin weight 0).

    This object is based on the TimeSeries object, but has many additional methods
    for manipulating modes.

    Parameters
    ----------
    input_array : (..., N, ..., M, ...) array_like
        Input data representing the dependent variable, in any form that can be
        converted to a numpy array.  This includes scalars, lists, lists of tuples,
        tuples, tuples of tuples, tuples of lists, and numpy ndarrays.  It can have
        an arbitrary number of dimensions, but the length `N` along `time_axis`
        must match the length of `time`, and the length `M` along `modes_axis` (see
        below) must be consistent with `ell_min` and `ell_max`.  Values must be
        finite.
    time : (N,) array_like
        1-D array containing values of the independent variable.  Values must be
        real, finite, and in strictly increasing order.
    time_axis : int, optional
        Axis along which `input_array` is assumed to be varying in time, meaning
        that for `time[i]` the corresponding values are `np.take(input_array, i,
        axis=time_axis)`.  If this is not given, the first axis of `input_array`
        that has the same length as `time` is chosen as the time axis — which may
        be prone to errors.
    modes_axis : int
        Axis of the array along which the modes are stored.  See Notes below.
    ell_min : int
        Smallest value of ℓ stored in the data
    ell_max : int
        Largest value of ℓ stored in the data

    Notes
    -----
    We assume that the modes at a given time, say, are stored in a flat array, in
    order of increasing `m` and `ell`, with `m` varying most rapidly — something
    like the following:

        [f(ell, m) for ell in range(ell_min, ell_max+1) for m in range(-ell,ell+1)]

    The total size is implicitly `ell_max * (ell_max + 2) - ell_min ** 2 + 1`.

    For backwards compatibility, it is possible to retrieve individual modes in the
    same way as the old NRAR-format HDF5 files would be read, as in

        h_22 = waveform["Y_l2_m2.dat"]

    Note that "History.txt" may not contain anything but an empty string, because
    history is not retained in more recent data formats.  Also note that — while
    not strictly a part of this class — the loaders that open waveform files will
    return a dict-like object when the extrapolation order is not specified.  That
    object can also be accessed in a backwards-compatible way much like the root
    directory of the NRAR-format HDF5 files.  For example:

        with sxs.loadcontext("rhOverM_Asymptotic_GeometricUnits_CoM.h5") as f:
            h_22 = f["Extrapolated_N2.dir/Y_l2_m2.dat"]

    This code is identical to the equivalent code using `h5py` except that the call
    to `h5py.File` is replaced with the call to `sxs.loadcontext`.  The `.dat`
    datasets are reconstructed on the fly, but should be bitwise-identical to the
    output from the HDF5 file whenever the underlying format is NRAR.

    """
    import functools

    def __new__(cls, input_array, *args, **kwargs):
        for requirement in ["modes_axis", "ell_min", "ell_max"]:
            if requirement not in kwargs and requirement not in getattr(input_array, "_metadata", {}):
                raise ValueError(f"{cls} could not find required argument '{requirement}'")
        self = super().__new__(cls, input_array, *args, **kwargs)
        return self

    def __getitem__(self, key):
        if isinstance(key, str):
            if key == "History.txt":
                return self._metadata.get("history", "")
            # Assume we're asking for Y_l2_m2.dat or something
            if self.data.shape != (self.n_times, self.n_modes):
                raise ValueError(f"Data has shape {self.data.shape}, which is incompatible with NRAR format")
            match = NRAR_mode_regex.match(key)
            if not match:
                raise ValueError(f"Key '{key}' did not match mode format")
            ell, m = int(match["L"]), int(match["M"])
            if ell < self.ell_min or ell > self.ell_max:
                raise ValueError(f"Key '{key}' requested ell={ell} value not found in this data")
            if abs(m) > ell:
                raise ValueError(f"Key '{key}' requested (ell, m)=({ell}, {m}) value does not make sense")
            index = self.index(ell, m)
            data = np.take(self.data, index, axis=self.modes_axis).view(float).reshape((-1, 2))
            return np.hstack((self.time[:, np.newaxis], data))
        obj, time_key = self._slice(key)
        if time_key is None:
            raise ValueError(f"Fancy indexing (with {key}) is not supported")
        obj = obj.view(type(self))
        if "frame" in obj._metadata and obj.frame.shape == (self.n_times, 4):
            obj._metadata["frame"] = obj.frame[time_key, :]
        return obj

    @property
    def modes_axis(self):
        """Axis of the array storing the various modes

        See the documentation of this class for an explanation of what this means.

        See Also
        --------
        time_axis : Axis of the array along which time varies

        """
        return self._metadata["modes_axis"]

    @property
    def ell_min(self):
        """Smallest value of ℓ stored in the data"""
        return self._metadata["ell_min"]

    @property
    def ell_max(self):
        """Largest value of ℓ stored in the data"""
        return self._metadata["ell_max"]

    @property
    def n_modes(self):
        """Total number of mode weights stored in the data"""
        return self.shape[self.modes_axis]

    @property
    def LM(self):
        """Array of (ell, m) values in the data

        This array is just a flat array of `[ell, m]` pairs.  It is automatically
        recomputed each time `ell_min` or `ell_max` is changed.  Specifically, it is

            np.array([
                [ell, m]
                for ell in range(self.ell_min, self.ell_max+1)
                for m in range(-ell, ell+1)
            ])

        """
        return np.array([[ell, m] for ell in range(self.ell_min, self.ell_max+1) for m in range(-ell, ell+1)])

    def index(self, ell, m):
        """Mode index of given (ell,m) mode in the data

        Parameters
        ----------
        ell : int
        m : int

        Returns
        -------
        idx : int
            Index such that self.LM[idx] == [ell, m]

        """
        return spherical.LM_index(ell, m, self.ell_min)

    @property
    def abs(self):
        """Absolute value of the data

        Returns
        -------
        absolute : TimeSeries
            Because the absolute values make no sense as mode weights, this is just a
            plain TimeSeries object.

        See Also
        --------
        arg

        """
        return np.abs(self.view(TimeSeries))

    @property
    def arg(self):
        """Complex phase angle of the data

        Note that the result is not "unwrapped", meaning that there may be
        discontinuities as the phase approaches ±π.

        Returns
        -------
        phase : TimeSeries
            Values are in the interval (-π, π].

        See Also
        --------
        numpy.angle
        arg_unwrapped

        """
        return np.angle(self.view(TimeSeries))

    @property
    def arg_unwrapped(self):
        """Complex phase angle of the data, unwrapped along the time axis

        The result is "unwrapped", meaning that discontinuities as the phase approaches
        ±π are removed by adding an appropriate amount to all following data points.

        Returns
        -------
        phase : TimeSeries
            Values at the initial time are in the interval (-π, π], but may evolve to
            arbitrary real values.

        See Also
        --------
        numpy.angle
        numpy.unwrap
        arg

        """
        return TimeSeries(np.unwrap(self.arg, axis=self.time_axis), self.time)

    @property
    def norm(self):
        """Compute the L² norm of the waveform

        Returns
        -------
        n : TimeSeries

        See Also
        --------
        numpy.linalg.norm
        numpy.take
        norm2 : squared version of this

        Notes
        -----
        The integral of the (squared) magnitude of the data equals the sum of the
        (squared) magnitude of the modes for orthonormal basis functions, meaning that
        the L² norm of the function equals the basic Euclidean norm of the modes.  We
        assume that these modes are expanded in a band-limited but otherwise complete
        orthonormal basis.

        """
        return TimeSeries(np.linalg.norm(self, axis=self.modes_axis), self.time)

    def max_norm_index(self, skip_fraction_of_data=4):
        """Index of time step with largest norm

        The optional argument skips a fraction of the data.  The default is 4, which
        means that it skips the first 1/4 of the data, and only searches the last 3/4
        of the data for the max.  This must be strictly greater than 1, or the entire
        data is searched for the maximum of the norm.

        """
        if skip_fraction_of_data <= 1:
            return np.argmax(self.norm)
        else:
            i = int(self.n_times // skip_fraction_of_data)
            return np.argmax(self[i:].norm) + i

    def max_norm_time(self, skip_fraction_of_data=4):
        """Return time at which largest norm occurs in data

        See `help(max_norm_index)` for explanation of the optional argument.

        """
        return self.t[self.max_norm_index(skip_fraction_of_data=skip_fraction_of_data)]

    def interpolate(self, new_time, derivative_order=0, out=None):
        """Interpolate this object to a new set of times

        Note that if this object has "frame" data and the derivative order is nonzero,
        it is not entirely clear what is desired.  In those cases, the frame is just
        interpolated to the new times, but no derivative or antiderivative is taken.

        Parameters
        ----------
        new_time : array_like
            Points to evaluate the interpolant at
        derivative_order : int, optional
            Order of derivative to evaluate.  If negative, the antiderivative is
            returned.  Default value of 0 returns the interpolated data without
            derivatives or antiderivatives.  Must be between -3 and 3, inclusive.

        See Also
        --------
        scipy.interpolate.CubicSpline :
            The function that this function is based on.
        antiderivative :
            Calls this funtion with `new_time=self.time` and
            `derivative_order=-antiderivative_order` (defaulting to a single
            antiderivative).
        derivative :
            Calls this function `new_time=self.time` and
            `derivative_order=derivative_order` (defaulting to a single derivative).
        dot :
            Property calling `self.derivative(1)`.
        ddot :
            Property calling `self.derivative(2)`.
        int :
            Property calling `self.antiderivative(1)`.
        iint :
            Property calling `self.antiderivative(2)`.

        """
        result = TimeSeries.interpolate(self, new_time, derivative_order=derivative_order, out=out)
        if self.frame.shape == (self.n_times, 4) and not np.array_equal(self.time, result.time):
            self._metadata["frame"] = quaternionic.squad(self.frame, self.time, result.time)
        return result

    def truncate(self, tol=1e-10):
        """Truncate the precision of this object's data in place

        This function sets bits in the data to 0 when they have lower significance than
        will alter the norm of the Waveform by a fraction `tol` at that instant in
        time.

        """
        if tol != 0.0:
            tol_per_mode = tol / np.sqrt(self.n_modes)
            abs_tolerance = np.linalg.norm(self.ndarray, axis=self.modes_axis, keepdims=True) * tol_per_mode
            super().truncate(abs_tolerance)

    def convert_to_conjugate_pairs(self):
        """Convert modes to conjugate-pair format in place

        This function alters this object's modes to store the sum and difference of
        pairs with opposite `m` values.  If we denote the modes `f[l, m]`, then we
        define

            s[l, m] = (f[l, m] + f̄[l, -m]) / √2
            d[l, m] = (f[l, m] - f̄[l, -m]) / √2

        For m<0 we replace the mode data with `d[l, -m]`, for m=0 we do nothing, and
        for m>0 we replace the mode data with `s[l, m]`.  That is, the mode data on
        output look like this:

            [d[2, 2], d[2, 1], f[2, 0], s[2, 1], s[2, 2], d[3, 3], d[3, 2], ...]

        The factor of √2 is chosen so that the norm (sum of the magnitudes squared) at
        each time for this data is the same as it is for the original data.

        """
        mode_plus = np.empty_like(self.ndarray[..., 0])
        mode_minus = np.empty_like(mode_plus)
        for ell in range(self.ell_min, self.ell_max + 1):
            for m in range(1, ell + 1):
                i_plus = self.index(ell, m)
                i_minus = self.index(ell, -m)
                mode_plus[:] = self.ndarray[..., i_plus]
                mode_minus[:] = self.ndarray[..., i_minus]
                self.ndarray[..., i_plus] = (mode_plus + np.conjugate(mode_minus)) / np.sqrt(2)
                self.ndarray[..., i_minus] = (mode_plus - np.conjugate(mode_minus)) / np.sqrt(2)

    def convert_from_conjugate_pairs(self):
        """Convert modes from conjugate-pair format in place

        This function reverses the effects of `convert_to_conjugate_pairs`.  See that
        function's docstring for details.

        """
        mode_plus = np.empty_like(self.ndarray[..., 0])
        mode_minus = np.empty_like(mode_plus)
        for ell in range(self.ell_min, self.ell_max + 1):
            for m in range(1, ell + 1):
                i_plus = self.index(ell, m)
                i_minus = self.index(ell, -m)
                mode_plus[:] = self.ndarray[..., i_plus]
                mode_minus[:] = self.ndarray[..., i_minus]
                self.ndarray[..., i_plus] = (mode_plus + mode_minus) / np.sqrt(2)
                self.ndarray[..., i_minus] = np.conjugate(mode_plus - mode_minus) / np.sqrt(2)

    def evaluate(self, *directions):
        """Evaluate waveform in a particular direction or set of directions

        Parameters
        ----------
        directions : array_like
            Directions are specified using the usual spherical coordinates, and an
            optional polarization angle (see Notes below).  These can be expressed as 2
            or 3 floats (where the third is the polarization angle), or as an array
            with final dimension of size 2 or 3.  Input arrays can have multiple
            leading dimensions; the final dimension is always considered to hold the
            directions, and the other dimensions are retained in the output.

        Returns
        -------
        signal : array_like
            Note that this is complex-valued, meaning that it represents both
            polarizations.  To get the signal measured by a single detector, just take
            the real part.

        Notes
        -----
        Here, we assume that the reference basis is the basis defining the
        spin-weighted spherical harmonics used to decompose the waveform into modes.
        Then, we define the spherical coordinates (θ, ϕ) such that θ is the polar angle
        (angle between the z axis and the point) and ϕ is the azimuthal angle (angle
        between x axis and orthogonal projection of the point into the x-y plane).
        This gives rise to the standard unit tangent vectors (θ̂, ϕ̂).

        We also define the polarization angle ψ as the rotation through which we must
        rotate the vector θ̂) in a positive sense about n̂ to line up with the vector
        defining the legs of the detector

        Examples
        --------
        We can evaluate the signal in a single direction:

        >>> θ, ϕ, ψ = 0.1, 0.2, 0.3
        >>> w.evaluate(θ, ϕ)  # Default polarization angle
        >>> w.evaluate(θ, ϕ, ψ)  # Specified polarization angle

        Or we can evaluate in a set of directions:

        >>> w.evaluate([[θ, ϕ], [θ+0.4, ϕ], [θ+0.8, ϕ]])

        We can also evaluate on a more extensive set of directions.  Here, we construct
        an equi-angular grid to evaluate the waveform on (though irregular grids are
        also acceptable as long as you can pack them into a numpy array).

        >>> n_theta = n_phi = 2 * w.ell_max + 1
        >>> equiangular = np.array([
            [
                [theta, phi]
                for phi in np.linspace(0.0, 2*np.pi, num=n_phi, endpoint=False)
            ]
            for theta in np.linspace(0.0, np.pi, num=n_theta, endpoint=True)
        ])
        >>> w.evaluate(equiangular)

        """
        import string
        if len(directions) == 1:
            directions = directions[0]
        directions = np.asarray(directions, dtype=float)
        if directions.shape[-1] == 2:
            R = quaternionic.array.from_spherical_coordinates(directions)
            #R = quaternionic.array.from_spherical_coordinates(directions[..., 0], directions[..., 1])
        elif directions.shape[-1] == 3:
            R = quaternionic.array.from_euler_angles(directions[..., 1], directions[..., 0], directions[..., 2])
        else:
            raise ValueError(f"Final dimension of input directions must have size 2 or 3, not {directions.shape[-1]}")
        # R.shape == directions.shape[:-1] + (4,)
        if self.frame == np.atleast_2d(quaternionic.one):
            sYlm = spherical.SWSH_grid(R, self.spin_weight, self.ell_max)
            sYlm = sYlm[..., spherical.LM_index(self.ell_min, -self.ell_min, 0):]  # Chop off leading zeros
            # sYlm.shape == directions.shape[:-1] + (self.n_modes,)
            result = np.tensordot(self.data, sYlm, axes=[[self.modes_axis], [-1]])
        else:
            # We have to account for a rotating frame
            frame = self.frame.reshape((self.frame.shape[0],) + (1,)*(R.ndim-1) + (4,))
            R = frame.inverse * R[np.newaxis]
            sYlm = spherical.SWSH_grid(R, self.spin_weight, self.ell_max)
            sYlm = sYlm[..., spherical.LM_index(self.ell_min, -self.ell_min, 0):]  # Chop off leading zeros
            # sYlm.shape == frame.shape[:1] + directions.shape[:-1] + (self.n_modes,)
            self_indices = list(string.ascii_letters[:self.ndim])
            self_indices[self.time_axis] = "y"
            self_indices[self.modes_axis] = "z"
            self_indices = ''.join(self_indices)
            sYlm_indices = list(string.ascii_letters[self.ndim:self.ndim+sYlm.ndim])
            sYlm_indices[0] = "y"
            sYlm_indices[-1] = "z"
            sYlm_indices = ''.join(sYlm_indices)
            result_indices = self_indices.replace("z", "") + sYlm_indices.replace("y", "").replace("z", "")
            subscripts = f"{self_indices},{sYlm_indices}->{result_indices}"
            result = np.einsum(subscripts, self.data, sYlm)
        return TimeSeries(result, self.time)

    # TODO:
    # expectation_value_LL
    # expectation_value_Ldt
    # angular_velocity
    # expectation_value_L
    # inner_product
    # # Don't bother with inner_product_LL, as it doesn't appear to be used; maybe a more general version?
    # mode_frame : Minimally rotating O'Shaughnessy et al. frame
    # to_mode_frame
    # corotating_frame

    def rotate(self, quat):
        """Rotate decomposition basis of modes represented by this waveform

        This returns a new waveform object, with missing "frame" data.

        Parameters
        ----------
        quat : quaternionic.array
            This must have the same number of quaternions as the number of times in the
            waveform.

        """
        if quat.shape != (self.n_times, 4):
            raise ValueError(f"Quaternionic array shape {quat.shape} not understood; expected {(self.n_times, 4)}")
        D = np.empty((spherical.WignerD._total_size_D_matrices(self.ell_min, self.ell_max),), dtype=complex)
        new_data = self.data.copy()
        quat2spinor = quat.two_spinor
        _rotate_decomposition_basis(new_data, quat2spinor.a, quat2spinor.b, self.ell_min, self.ell_max, D)
        new_metadata = self._metadata.copy()
        new_metadata.pop("frame")
        w = type(self)(new_data, **new_metadata)
        return w

    def to_inertial_frame(self):
        """Return a copy of this waveform in the inertial frame"""
        if "frame" not in self._metadata:
            raise ValueError("This waveform has no frame information")
        if self.frame.shape[0] == 1:
            raise ValueError("This waveform appears to already be in an inertial frame")
        if self.frame.shape != (self.n_times, 4):
            raise ValueError(f"Frame shape {frame.shape} not understood; expected {(self.n_times, 4)}")
        w = self.rotate(~self.frame)
        w._metadata["frame_type"] = "inertial"
        return w

    def to_corotating_frame(
        self, R0=None, tolerance=1e-12, z_alignment_region=None, return_omega=False, truncate_log_frame=False
    ):
        """Return a copy of this waveform in the corotating frame

        The corotating frame is defined to be a rotating frame for which the (L² norm
        of the) time-dependence of the modes expressed in that frame is minimized.
        This leaves the frame determined only up to an overall rotation.  In this 

        Parameters
        ----------
        R0 : quaternionic, optional
            Initial value of frame when integrating angular velocity.  Defaults to the
            identity.
        tolerance : float, optional
            Absolute tolerance used in integration of angular velocity
        z_alignment_region : {None, 2-tuple of floats}, optional
            If not None, the dominant eigenvector of the <LL> matrix is aligned with
            the z axis, averaging over this portion of the data.  The first and second
            elements of the input are considered fractions of the inspiral at which to
            begin and end the average.  For example, (0.1, 0.9) would lead to starting
            10% of the time from the first time step to the max norm time, and ending
            at 90% of that time.
        return_omega : bool, optional
            If True, return a 2-tuple consisting of the waveform in the corotating
            frame (the usual returned object) and the angular-velocity data.  That is
            frequently also needed, so this is just a more efficient way of getting the
            data.
        truncate_log_frame : bool, optional
            If True, set bits of log(frame) with lower significance than `tolerance` to
            zero, and use exp(truncated(log(frame))) to rotate the waveform.  Also
            returns `log_frame` along with the waveform (and optionally `omega`)

        """
        raise NotImplementedError()
        frame, omega = corotating_frame(
            self, R0=R0, tolerance=tolerance, z_alignment_region=z_alignment_region, return_omega=True
        )
        if truncate_log_frame:
            log_frame = np.log(frame).ndarray
            power_of_2 = 2 ** int(-np.floor(np.log2(2 * tolerance)))
            log_frame = np.round(log_frame * power_of_2) / power_of_2
            frame = np.exp(quaternionic.array(log_frame))
        w = self.rotate_decomposition_basis(frame)
        w._metadata["frame_type"] = "corotating"
        w._metadata["frame"] = frame
        if return_omega:
            if truncate_log_frame:
                return (w, omega, log_frame)
            else:
                return (w, omega)
        else:
            if truncate_log_frame:
                return (w, log_frame)
            else:
                return w


@jit("void(c16[:,:], c16[:], c16[:], i8, i8, c16[:])")
def _rotate_decomposition_basis(data, R_basis_a, R_basis_b, ell_min, ell_max, D):
    """Rotate data by a different rotor at each point in time

    `D` is just a workspace, which holds the Wigner D matrices.
    During the summation, it is also used as temporary storage to hold
    the results for each item of data, where the first row in the
    matrix is overwritten with the new sums.

    """
    for i_t in range(data.shape[0]):
        spherical._Wigner_D_matrices(R_basis_a[i_t], R_basis_b[i_t], ell_min, ell_max, D)
        for ell in range(ell_min, ell_max + 1):
            i_data = ell ** 2 - ell_min ** 2
            i_D = spherical._linear_matrix_offset(ell, ell_min)

            for i_m in range(2 * ell + 1):
                new_data_mp = 0j
                for i_mp in range(2 * ell + 1):
                    new_data_mp += data[i_t, i_data + i_mp] * D[i_D + i_m + (2 * ell + 1) * i_mp]
                D[i_D + i_m] = new_data_mp
            for i_m in range(2 * ell + 1):
                data[i_t, i_data + i_m] = D[i_D + i_m]



"""Greedy algorithm"""

def minimal_grid(x, y, tol=1e-6, rel=False):
    """Greedy algorithm for building a reduced-order spline

    This function is a slight generalization of the core function from `romspline`
    -- namely, `romspline.greedy._greedy` -- tailored to the needs of this package.
    This function should behave nearly the same as the `romspline` function (except
    that this function returns fewer quantities), but also accepts a callable
    object as the `tol` parameter, to enable more general evaluations of the error.

    Parameters
    ----------
    x : array of float
        An ordered series of sampling points.
    y : array of float
        Function values sampled at points given by `x`.
    tol : {float, callable}, optional
        If given a float, this function will iteratively increase the number of
        sampling points until the spline formed from those points is within this
        value of the input data at each input sample.  If given a callable object,
        that object must be as described in the Notes section below.  The default
        value is 1e-6.
    rel : bool, optional
        If `tol` is a float and this is True, the tolerance will be taken relative
        to the maximum absolute value in the input `y` data.  If False (the
        default), the tolerance will be treated as an absolute value.  Note that if
        `tol` is callable this parameter is ignored.

    Returns
    -------
    include_sample : array of bool
        Boolean array of the same length as `x` determining whether or not the
        corresponding point should be included in a reduced-order spline.  Note
        that numpy arrays can be indexed with this array, so that
        `x[include_sample]` will give the set of knots and `y[include_sample]` will
        give the corresponding set of data values with which to build the
        reduced-order spline.

    Notes
    -----
    The `tol` callable, if given, is used to decide whether or not the current
    trial spline is acceptable, and if not which sample point to include for the
    next trial spline.  It must be callable with the signature (x, y, y_greedy),
    and is expected to return either None (if the trial is acceptable) or the index
    of the next sample point to include.  Here, `x` and `y` are just the
    corresponding inputs to this function, and `y_greedy` is a set of data points
    representing the values of the trial reduced-order spline evaluated on `x`.
    For example, the following function would behave equivalently to the default
    parameters:

        def tol(x, y, y_greedy):
            errors = np.abs(y-y_greedy)
            i_next = np.argmax(errors)
            if errors[i_next] < 1e-6:
                return None
            return i_next

    """
    import numpy as np
    from scipy.interpolate import CubicSpline as spline

    deg = 3  # This used to be a parameter, before switching to CubicSpline

    if callable(tol):
        next_sample = tol
    else:
        if rel:
            # Rescale the tolerance
            ymax = np.max(np.abs(y))
            if ymax == 0.0:
                raise ValueError("All input `y` data samples are zero.")
            tol *= ymax

        # Construct the function used to evaluate the greedy spline
        def next_sample(x, y, y_greedy):
            errors = np.abs(y-y_greedy)
            i_next = np.argmax(errors)
            if errors[i_next] < tol:
                return None
            return i_next

    # Create an array describing whether or not to include a given
    # sample in the spline.  Start with all False.
    include_sample = np.zeros(len(x), dtype=bool)

    # Initialize greedy algorithm with evenly spaced indices, as many
    # as needed for the desired spline degree.
    include_sample[np.linspace(0, len(x)-1, num=deg+1, dtype=int)] = True

    # Greedy algorithm
    for _ in range(len(x)):
        # Spline interpolant on current set of knots
        s = spline(x[include_sample], y[include_sample])

        # Evaluate this spline
        i_next = next_sample(x, y, s(x))

        # Break out of this loop if `tol` is satisfied
        if i_next is None:
            break

        # Include data point that gives largest interpolation error
        include_sample[i_next] = True

    return include_sample



import contextlib
import pathlib
import tempfile
import numpy as np
import pytest
import sxs

file_name = "SXS:BBH:0004/Lev6/Horizons.h5"


def test_spec_format():
    """Ensure that we can slice the output of spec_horizons_h5 just like the original Horizons.h5"""
    import h5py

    quantities = [
        "ArealMass", "ChristodoulouMass", "CoordCenterInertial", "DimensionfulInertialSpin",
        "DimensionfulInertialSpinMag", "chiInertial", "chiMagInertial"
    ]

    with contextlib.redirect_stdout(None):
        catalog = sxs.load("catalog")
    selected = catalog.select_files(file_name)
    selected_path = sxs.utilities.sxs_path_to_system_path(list(selected.values())[0]["truepath"])

    with contextlib.redirect_stdout(None):
        horizons = sxs.load(file_name)
    cached_path = sxs.sxs_directory("cache") / selected_path

    with h5py.File(cached_path, "r") as f:
        for horizon in "ABC":
            for quantity in quantities:
                name = f"Ah{horizon}.dir/{quantity}.dat"
                a, b = f[name], horizons[name]
                if 'Mag' in quantity:
                    assert np.allclose(a, b, atol=1e-11, rtol=1e-11)
                else:
                    assert np.array_equal(a, b)


def test_xmb_format():
    # horizons_spec = sxs.horizons.spec_horizons_h5.load(file_name)
    with contextlib.redirect_stdout(None):
        horizons_spec = sxs.load(file_name)

    with tempfile.TemporaryDirectory() as temp_dir:
        file = pathlib.Path(temp_dir) / 'horizons.h5'
        sxs.horizons.xor_multishuffle_bzip2.save(horizons_spec, file)
        with pytest.raises(ValueError):
            horizons_error = sxs.horizons.spec_horizons_h5.load(file)
        horizons_xmb = sxs.horizons.xor_multishuffle_bzip2.load(file)

    for horizon_name in "ABC":
        h_spec = horizons_spec[horizon_name]
        h_xmb = horizons_xmb[horizon_name]
        for attr in [a for a in dir(h_spec) if not a.startswith('_')]:
            d_spec = getattr(h_spec, attr)
            d_xmb = getattr(h_xmb, attr)
            assert np.array_equal(d_spec, d_xmb)



"""Functions to convert waveform quantities to LVC format"""

import numpy as np
import h5py


def convert_modes(sxs_format_waveform, metadata, out_filename,
                  modes, extrapolation_order="Extrapolated_N2",
                  log=print, truncation_time=None,
                  tolerance=5e-07, truncation_tol=None):
    """Computes amplitude and phase for an SXS-format waveform, computes
    ROM-spline for each mode, and writes to file.
    """
    from time import perf_counter
    from . import WaveformAmpPhase, Dataset

    extrap = str(extrapolation_order) + ".dir"

    if truncation_time is None:
        truncation_time = metadata['reference_time'] / (metadata["reference_mass1"] + metadata["reference_mass2"])

    mode_strings = ["Y_l{0[0]}_m{0[1]}.dat".format(ellm) for ellm in modes]
    h = WaveformAmpPhase.read(sxs_format_waveform, extrap, mode_strings=mode_strings, start_time=truncation_time)

    start_time = h.t[0]

    peak_time = h.t[h.peak_index()]

    t = h.t - peak_time

    if truncation_tol is True:  # Test for actual identity, not just equality
        truncation_tol = 5e-2 * tolerance
    elif truncation_tol is False:
        truncation_tol = None

    with h5py.File(out_filename, 'w') as out_file:
        out_file.create_dataset('NRtimes', data=t)
        for i, mode in enumerate(modes):
            amp = h.data[:, 2*i]
            phase = h.data[:, 2*i+1]

            phase_out = Dataset.from_data(t, phase, tolerance, error_scaling=amp, truncation_tol=truncation_tol)
            amp_out = Dataset.from_data(t, amp, tolerance, truncation_tol=truncation_tol)

            out_group_amp = out_file.create_group('amp_l{0[0]}_m{0[1]}'.format(mode))
            amp_out.write(out_group_amp)
            out_group_phase = out_file.create_group('phase_l{0[0]}_m{0[1]}'.format(mode))
            phase_out.write(out_group_phase)

    return start_time, peak_time, h.version_hist



"""Functions to convert metadata to the LVC format"""

import time
import json
import numpy as np
import h5py

# Compute 1 solar mass * G/c^3
msun_seconds = 4.925491025543576e-06  # LAL value


def sxs_id_from_alt_names(alt_names):
    """Takes an array of alternative names from an SXS metadata.json file
    and returns the SXS ID of the simulation."""
    pattern = 'SXS'
    if not isinstance(alt_names, (list, tuple)):
        alt_names = [alt_names]
    sxs_id = str(next((ss for ss in alt_names if pattern in ss), None))
    return sxs_id


def simulation_type_from_spins(dimensionless_spin_1, dimensionless_spin_2):
    """Classify simulation as non-spinning, aligned-spins, or precessing

    Classifies a simulation with dimensionless_spin_1 on the primary
    (larger) black hole and dimensionless_spin_2 on the secondary black
    hole as nonspinning (if no spin component > 0.01), aligned-spin (if
    only z components are > 0.01), or precessing (otherwise).

    """
    spin_zero_threshold = 0.01  # treat spins smaller than this as zero
    # Types defined in arXiv:1703.01076
    nonspinning_type = "non-spinning"
    aligned_type = "aligned-spins"
    precessing_type = "precessing"

    simulation_type = nonspinning_type

    if (abs(dimensionless_spin_1[2]) > spin_zero_threshold or
            abs(dimensionless_spin_2[2]) > spin_zero_threshold):
        simulation_type = aligned_type

    if (abs(dimensionless_spin_1[0]) > spin_zero_threshold or
            abs(dimensionless_spin_2[0]) > spin_zero_threshold or
            abs(dimensionless_spin_1[1]) > spin_zero_threshold or
            abs(dimensionless_spin_2[1]) > spin_zero_threshold):
        simulation_type = precessing_type

    return simulation_type


def find_comparable_simulations(sxs_id, catalog, catalog_resolutions):
    """Return LVC-format filename containing comparable simulation

    Given an SXD ID sxs_id, a dictionary catalog_data (whose keys are
    SXS ID numbers and whose values are dictionaries including masses
    and spins) and a dictionary catalog_resolutions (whose keys are SXS
    ID numbers and whose values are dictionaries containing lists of
    integers indicating what resolutions are available in the catalog
    for that SXS ID number), return an LVC-format filename that is a
    comparable simulation with multiple resolutions available.

    """
    mass1 = catalog[sxs_id]['initial_mass1']
    mass2 = catalog[sxs_id]['initial_mass2']
    spin1 = catalog[sxs_id]['initial_dimensionless_spin1']
    spin2 = catalog[sxs_id]['initial_dimensionless_spin2']
    mass_ratio = mass1 / mass2
    spin1_magnitude = np.linalg.norm(spin1)
    spin2_magnitude = np.linalg.norm(spin2)

    # Select SXS ID numbers with multiple resolutions available
    has_multiple_resolutions = []
    for key in catalog_resolutions:
        if len(catalog_resolutions[key]) > 1:
            has_multiple_resolutions.append(key)

    # Select SXS ID numbers with multiple resolutions and the same
    # initial data type and spec revision
    same_id_and_spec_revision = []
    for key in has_multiple_resolutions:
        if (catalog[key]['spec_revisions'] == catalog[sxs_id]['spec_revisions']
                and catalog[key]['initial_data_type'] == catalog[sxs_id]['initial_data_type']):
            same_id_and_spec_revision.append(key)

    if len(same_id_and_spec_revision) > 0:
        has_multiple_resolutions = same_id_and_spec_revision

    # Select an SXS ID number with multiple resolutions that is "similar" to
    # sxs_id. First, try to find a case with the same initial data type and
    # SpEC revision. If that is possible, then choose from among the ones with
    # the closest mass ratio and spin magnitudes.  Otherwise, just select a
    # simulation with the closest mass ratios and spin magnitudes.
    #
    # Note: as in the previous script, here I use initial masses and spins, not
    # relaxed masses and spins.
    mass_spin_diff_best = np.inf
    key_best = has_multiple_resolutions[0]
    for key in has_multiple_resolutions:
        current_mass1 = catalog[key]['initial_mass1']
        current_mass2 = catalog[key]['initial_mass2']
        current_spin1 = catalog[key]['initial_dimensionless_spin1']
        current_spin2 = catalog[key]['initial_dimensionless_spin2']
        current_mass_ratio = current_mass1 / current_mass2
        current_spin1_magnitude = np.linalg.norm(current_spin1)
        current_spin2_magnitude = np.linalg.norm(current_spin2)
        mass_spin_diff = (
                np.abs(mass_ratio - current_mass_ratio)
                + np.abs(spin1_magnitude - current_spin1_magnitude)
                + np.abs(spin2_magnitude - current_spin2_magnitude)
        )
        if mass_spin_diff < mass_spin_diff_best:
            mass_spin_diff_best = mass_spin_diff
            key_best = key

    resolution_best = np.max(catalog_resolutions[key_best])
    return key_best.replace(':', '_') + "_Res" + str(resolution_best) + ".h5"


def write_metadata_from_sxs(out_filename, resolution, metadata, catalog,
                            catalog_resolutions, start_time, peak_time, l_max, log=print):
    """Write metadata to an LVC-format file.

    Adds both auxiliary-info/metadata.json and setting attributes
    conforming to the format given by arXiv:1703.01076. Input
    arguments are the output filename out_filename, the resolution of
    the simulation (an int), metadata (read from metadata.json) for
    this simulation/resolution, the start_time, peak time, and l_max
    determined by waveforms.convert_modes.  The argument
    catalog_resolutions is a dictionary (read from a json file) whose
    keys are SXS ID numbers, and whose values are lists of integers
    corresponding to the resolutions available for that SXS ID number
    in the SXS catalog. The argument catalog should point to the SXS
    catalog metadata JSON file, available at
    https://data.black-holes.org/catalog.json or via
    get_sxs_public_metadata.py.

    """
    log("Writing metadata")
    # Get the SXS ID for the current simulation
    names = metadata['alternative_names']
    if not isinstance(names, (list, tuple)):
        names = [names]
    sxs_id = sxs_id_from_alt_names(names)

    with h5py.File(out_filename, 'a') as out_file:

        # Put the metadata for this simulation into the auxiliary-info group
        aux_group = out_file.create_group('auxiliary-info')
        json_this_sim = json.dumps(metadata)
        aux_group.create_dataset('metadata.json', data=json_this_sim)

        # Note: all numerical quantities for LVC attributes
        # are at the reference time unless otherwise indicated
        mass1 = metadata["reference_mass1"]
        mass2 = metadata["reference_mass2"]
        total_mass = mass1 + mass2
        eta = (mass1 * mass2) / np.square(total_mass)

        # Round slightly larger mass ratios to 1
        # CHECK ME: should we do this?
        if 0.25 < eta < 0.2501:
            eta = 0.25

        dimensionless_spin1 = metadata["reference_dimensionless_spin1"]
        dimensionless_spin2 = metadata["reference_dimensionless_spin2"]

        position1 = np.array(metadata["reference_position1"])
        position2 = np.array(metadata["reference_position2"])
        separation = position1 - position2

        n_hat = separation / np.linalg.norm(separation)
        omega_orbit_vec = metadata["reference_orbital_frequency"]
        omega_orbit = np.linalg.norm(omega_orbit_vec)
        omega_grav_22 = 2.0 * omega_orbit

        # Unit vector in direction of orbital angular momentum, which
        # is the same as the direction of the orbital angular velocity
        # at the reference time
        ln_hat = omega_orbit_vec / omega_orbit

        # Eccentricity quantities
        eccentricity = metadata['reference_eccentricity']
        if str(eccentricity)[0] == "<":
            eccentricity = float(str(eccentricity)[1:])
        if str(eccentricity)[0] == ">":
            eccentricity = float(str(eccentricity)[1:])
        elif str(eccentricity) == '[simulation too short]' or str(eccentricity) == '[unknown]':
            # CHECK ME: is this the right thing to do for cases where we can't
            # measure eccentricity?
            log("Warning: eccentricity not measured for this simulation")
            eccentricity = -1.0
        else:
            eccentricity = float(eccentricity)

        mean_anomaly = metadata['reference_mean_anomaly']
        if isinstance(mean_anomaly, str):
            if mean_anomaly == '[unknown]':
                mean_anomaly = -1.0
            else:
                mean_anomaly = float(mean_anomaly)

        # Set metadata attributes of the output file
        out_file.attrs['NR-group'] = "SXS"
        out_file.attrs['name'] = sxs_id
        out_file.attrs['alternative-names'] = ",".join(names)
        out_file.attrs['simulation-type'] \
            = simulation_type_from_spins(dimensionless_spin1, dimensionless_spin2)

        # our metdata uses lowercase for object types (bh, ns), but LVC wants
        # BH or NS, so convert to uppercase with upper()
        out_file.attrs['object1'] = metadata['object1'].upper()
        out_file.attrs['object2'] = metadata['object2'].upper()

        out_file.attrs['PN_approximant'] = "none, NR only"

        out_file.attrs['NR_reference_time'] = float()

        out_file.attrs['mass1'] = mass1
        out_file.attrs['mass2'] = mass2
        out_file.attrs['eta'] = eta

        # Metadata not in arXiv:1703.01076
        out_file.attrs['NR_reference_time'] = metadata['reference_time']
        out_file.attrs['NR_start_time'] = start_time
        out_file.attrs['NR_peak_time'] = peak_time  # not in Patricia's script
        out_file.attrs['NR_frame'] = 'inertial'

        out_file.attrs['f_lower_at_1MSUN'] = omega_grav_22 / (2.0 * np.pi * msun_seconds)
        out_file.attrs['Omega'] = omega_orbit
        out_file.attrs['spin1x'] = dimensionless_spin1[0]
        out_file.attrs['spin1y'] = dimensionless_spin1[1]
        out_file.attrs['spin1z'] = dimensionless_spin1[2]
        out_file.attrs['spin2x'] = dimensionless_spin2[0]
        out_file.attrs['spin2y'] = dimensionless_spin2[1]
        out_file.attrs['spin2z'] = dimensionless_spin2[2]
        out_file.attrs['nhatx'] = n_hat[0]
        out_file.attrs['nhaty'] = n_hat[1]
        out_file.attrs['nhatz'] = n_hat[2]
        out_file.attrs['LNhatx'] = ln_hat[0]
        out_file.attrs['LNhaty'] = ln_hat[1]
        out_file.attrs['LNhatz'] = ln_hat[2]

        out_file.attrs['eccentricity'] = eccentricity
        out_file.attrs['mean_anomaly'] = mean_anomaly

        out_file.attrs['type'] = "NRinjection"
        out_file.attrs['Format'] = 3
        out_file.attrs['NR-code'] = "SpEC"
        out_file.attrs['modification-date'] = time.strftime("%Y-%m-%d")
        out_file.attrs['point-of-contact-email'] = 'questions@black-holes.org'
        # CHECK ME: is this always the right reference to cite?
        out_file.attrs['INSPIRE-bibtex-keys'] = "Boyle:2019kee"

        # CHECK-ME: should this be public?
        out_file.attrs['license'] = "LVC-internal"

        out_file.attrs['NR-techniques'] = (
            "Quasi-Equilibrium-ID, GH, RWZ-h, "
            "Extrapolated-Waveform, ApproxKillingVectorSpin, Christodoulou-Mass"
        )
        out_file.attrs['Lmax'] = l_max

        # If highest resoultion available, 'production-run' = 1, otherwise
        # 'production-run' = 0. If multiple resolutions available, list their
        # file names for 'files-in-error-series'. Otherwise, list filenames of
        # a comparable simulation with multiple resolutions.
        resolutions = catalog_resolutions.get(sxs_id, [])
        if len(resolutions) > 1:
            error_name_base = out_filename.split('/')[-1].split("_Res")[0] + "_Res"
            error_series = [error_name_base + str(i) + ".h5" for i in resolutions]
            out_file.attrs['files-in-error-series'] = ",".join(error_series)
            out_file.attrs['comparable-simulation'] = ""
            if resolution == np.max(resolutions):
                out_file.attrs['production-run'] = 1
            else:
                out_file.attrs['production-run'] = 0
        else:
            out_file.attrs['production-run'] = 1
            out_file.attrs['files-in-error-series'] = ""
            out_file.attrs['comparable-simulation'] = ""




class Records(object):

    @classmethod
    def search(cls, q=None, sort=None, page=1, size=1000, sandbox=False, all_versions=False, max_pages=10):
        """Search public records

        It is possible to filter the results use the optional parameters.  Note that the web interface
        can sometimes be used to find search parameters by looking in the `search-hidden-params`
        parameter of the `invenio-search` tag.

        Example queries
        ---------------
        'title:"SXS:BBH:0003"'  # Finds titles with given string; use quotes for robustness
        'communities:sxs'  # Records in the 'sxs' Zenodo community
        'provisional_communities:sxs'  # Records awaiting approval by the community curator
        'owners: 38418'  # Find records by id number of owner

        Optional parameters
        -------------------
        q: string
            Search query, using Elasticsearch query string syntax.  See
            https://help.zenodo.org/guides/search/ for details.
        sort: string
            Sort order ('bestmatch' or 'mostrecent').  Prefix with minus to change from ascending to
            descending (e.g., '-mostrecent').
        page: int
            Page number for pagination
        size: int
            Number of results to return per page.  Note that Zenodo (as of this writing) seems to
            place a hard limit of 9999 responses.  Anything more will result in an error.  Use
            multiple pages to get more results.
        sandbox: bool [defaults to False]
            If True use sandbox.zenodo.org instead of the standard site.
        all_versions: bool [defaults to False]
            If True return all records, including older versions of published records.
        max_pages: int [defaults to 10]
            If the query returns a number of records equal to `size`, it is evidently incomplete.
            This function will attempt to retrieve successive pages until the number of records is
            less than `size`.  If the query is still incomplete after this many pages, just return
            what we've got.

        """
        import requests
        from . import url_sandbox, url_standard

        if sandbox:
            base_url = url_sandbox
        else:
            base_url = url_standard
        url = base_url + "api/records/"

        params={}
        if q is not None:
            params['q'] = q
        if sort is not None:
            params['sort'] = sort
        params['page'] = page
        params['size'] = size
        if all_versions:
            params['all_versions'] = ''

        r = requests.get(url, params=params, headers={'Accept': 'application/json'})
        if r.status_code != 200:
            print('An unknown error occurred when trying to access {0}.'.format(url))
            print('The search parameters were "{0}"'.format(params))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error

        json = r.json()
        if len(json) == size:
            page += 1
            if page > max_pages:
                print('Search is not yet complete after {0} pages; returning with what we have.'.format(max_pages))
                return r.json()
            return json + cls.search(q=q, sort=sort, page=page, size=size, sandbox=sandbox, all_versions=all_versions, max_pages=max_pages)

        return json



import contextlib
import pathlib
import tempfile
import numpy as np
import h5py
import pytest
import sxs

# see test_utilities.py for explanation
shortest_h_com_file = "SXS:BBH:0156v1/Lev5/rhOverM_Asymptotic_GeometricUnits_CoM.h5"


def test_backwards_compatibility():
    path = sxs.sxs_directory("cache") / sxs.utilities.sxs_path_to_system_path(shortest_h_com_file)
    with contextlib.redirect_stdout(None):
        with sxs.loadcontext(shortest_h_com_file, extrapolation_order=...) as h:
            with h5py.File(path, "r") as f:
                for group in f:
                    if group == "VersionHist.ver":
                        continue
                    g = f[group]
                    for d in g:
                        if d == "History.txt":
                            continue
                        assert np.array_equal(g[d], h[group][d])
                for group in f:
                    if group != "VersionHist.ver":
                        continue
                    assert np.array_equal(f[group], h[group])



"""Utilities that inherit from dict"""

class KeyPassingDict(dict):
    """Subclass of dict that enables extended keys

    When indexing this object, the key is split into two pieces: the part preceding
    the first "/" and everything after it.  The second part is passed to the
    __getitem__ method of whatever is returned by indexing this object with the
    first part of the key.  This is helpful when trying to emulate the behavior of
    HDF5 files, for example.

    """
    def __getitem__(self, key):
        parts = key.split("/", 1)
        value = super().__getitem__(parts[0])
        if len(parts) > 1:
            return value[parts[1]]
        else:
            return value
